const std = @import("std");

const msg_queue = @import("../msg_queue/root.zig");
const models = @import("../models/root.zig");

const Self = @This();

const log = std.log.scoped(.block_update_dispatcher);

/// Block update dispatcher config
pub const Config = struct {
    /// Queue capacity
    capacity: usize,
};

allocator: std.mem.Allocator,

/// Dispatcher config
config: *const Config,

/// Block update queue
queue: *msg_queue.MsgQueueUnmanaged(models.BlockUpdateMsg),

/// Data server
server: *std.net.Server,

pub fn init(
    allocator: std.mem.Allocator,
    config: *const Config,
    queue: *msg_queue.MsgQueueUnmanaged(models.BlockUpdateMsg),
    server: *std.net.Server,
) Self {
    return .{
        .allocator = allocator,
        .config = config,
        .queue = queue,
        .server = server,
    };
}

/// Sends one block update payload to a stream, terminates it with a newline.
fn send(stream: std.net.Stream, update: *const models.BlockUpdateMsg) !void {
    var buf: [models.BlockUpdateMsg.csv_max_len]u8 = undefined;
    const csv = try update.toCsv(&buf);
    stream.writeAll(csv) catch return;
    stream.writeAll(&.{'\n'}) catch return;
}

fn streamListener(self: *Self, stream: std.net.Stream, running: *std.atomic.Value(bool)) void {
    var buf: [512]u8 = undefined;
    while (running.load(.monotonic)) {
        const n = stream.read(&buf) catch 0;
        std.log.info("n: {}", .{n});
        if (n < buf.len) break;
    }

    std.log.info("terminating stream listener", .{});

    running.store(false, .release);
    self.queue.cond.broadcast();
}

pub fn start(self: *Self) !void {
    while (true) {
        const conn = try self.server.accept();
        log.info("Client connected", .{});
        defer conn.stream.close();

        // buffer for `tryDequeueAll` with maximum queue capacity as size
        const buf = try self.allocator.alloc(models.BlockUpdateMsg, self.queue.capacity());
        defer self.allocator.free(buf);

        var running = std.atomic.Value(bool).init(true);
        defer running.store(false, .release);
        _ = try std.Thread.spawn(.{}, streamListener, .{ self, conn.stream, &running });

        blk: {
            while (true) {
                const updates = try self.queue.dequeueAll(buf);
                if (!running.load(.monotonic)) break :blk;
                for (updates) |*update| {
                    if (running.load(.monotonic)) {
                        try send(conn.stream, update);
                    } else {
                        break :blk;
                    }
                }
            }
        }
    }
}

// /// Sends one block update payload to the stream, terminates it with a newline.
// fn send(self: *Self, update: *const models.BlockUpdateMsg) !void {
//     var buf: [models.BlockUpdateMsg.csv_max_len]u8 = undefined;
//     const stream = self.stream.?;
//     const csv = try update.toCsv(&buf);
//     try stream.writeAll(csv);
//     try stream.writeAll(&.{'\n'});
// }

// /// Starts dispatcher. Receives new queue entries and sends them to the stream.
// pub fn start(self: *Self) !void {
//     // buffer for `tryDequeueAll` with maximum queue capacity as size
//     const buf = try self.allocator.alloc(models.BlockUpdateMsg, self.queue.capacity());
//     defer self.allocator.free(buf);

//     while (true) {
//         const updates = try self.queue.tryDequeueAll(buf);
//         for (updates) |*update| try self.send(update);
//     }
// }
